#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "non_gui_defs.h"
#include "racing.h"

/* calculate the time a dog needs to complete 1, 2 or 3 laps,
   respectively. (one lap is 300 m long)
   depends on dog.speed and dog.stamina, naturally */
gfloat calculate_time(dog *Dog, gint laps)
{
    gint i;
    gfloat time, fade, first_lap;
    gfloat lap_skill;

    gfloat speed = sum(Dog->speed,0,98) * Dog->fitness;
    gfloat stamina = sum(Dog->stamina,0,98) * Dog->fitness;
    
    /* we combine the two basic values to a new one that determines
       the race result; the weights corresponding to the number of
       laps seem quite logical to me */
    if(laps == 1)
	lap_skill = 0.75 * speed + 0.25 * stamina;
    else if(laps == 2)
	lap_skill = 0.5 * speed + 0.5 * stamina;
    else
	lap_skill = 0.25 * speed + 0.75 * stamina;

    /* first calculate the basic time needed for the first lap
       depending on the dog's speed value.
       a superior dog in optimal shape maintains a speed
       of 65 km/h in a one-lap race. */
    first_lap = -0.03038 * lap_skill + 19.6076;

    /* now we randomise with the normal distribution with
       mean 'first_lap' and a 99.7% probability of the new first_lap
       being within a 3%-range of 'first_lap'.
       the two cutoff values 7 and 30 are needed only theretically */
    first_lap = gaussvt( 7, 0.97 * first_lap, 1.03 * first_lap, 30 );
    
    time = first_lap;
    
    /* if this isn't a sprint, calculate the fading factor
       which determines the time loss on lap 2 and 3 compared
       with lap 1. like time it's a linear function, in this case,
       naturally, of 'stamina':
       a dog with maximal stamina is expected to be 3% slower each
       additional lap, whereas a dog with really bad stamina loses
       10% */
    if(laps > 1)
    {
	fade = -0.000886 * stamina + 1.11772;

	/* the time for the first lap we have
	   already, so 'i' starts at 1 */
	for(i=1;i<laps;i++)
	    time += first_lap * pow( fade, i );
    }
    
    return time;
}

/* select randomly the dogs for the 12 races (2 per dog level) that
   take place each month; select randomly number of laps for each race.
   the index '9' is reserved for the human player's dog */
gint write_races(void)
{
    gint i,j;

    for(i=0;i<12;i++)
    {
	races[1][i].level = (gint)((gfloat)i / 2);
	races[1][i].laps = (gint)rint( rnd(1,3) );
	
	for(j=0;j<10;j++)
	{
	    races[1][i].times[j] = 200;
	    races[1][i].dogs[j] = NULL;
	    races[1][i].ranks[j] = j;
	}

	/* fill in dogs */
	select_dogs(i);
    }
    
    return 0;  
}

/* fill dogs into a race respecting the preferences (sprint/long distance)
   of the dogs */
gint select_dogs(gint race_number)
{
    gint i;
    gint order[120];
    gint number_of_dogs_on_level = 0;
    gint number_of_dogs_entered = 0;
    gint dogs_on_level[120];
    
    /* find out the ids of the dogs on the wanted level */
    for(i=20;i<400;i++)
    {
	if(dogs[i].level == races[1][ race_number ].level &&
	   number_of_dogs_on_level < 120)
	{
	    /* we have to filter out the dogs already entered
	       into the previous race if this is the second race
	       on one level */
	    if( race_number % 2 == 0 ||
		(race_number % 2 == 1 &&
		 entered_in_races(dogs[i]) == 0) )
	    {
		dogs_on_level[ number_of_dogs_on_level ] = i;
		number_of_dogs_on_level++;
	    }
	}
    }

    /* write a permutation to ensure random selection afterwards */
    write_permutation(order, 0, number_of_dogs_on_level - 1);

    /* enter the dogs into race */
    for(i=0;i<number_of_dogs_on_level;i++)
    {
	/* respect preference if it's a sprgint or a
	   long distance race; only enter healthy dogs */
	if( preference_of(dogs[ dogs_on_level[ order[i] ] ])
	    == races[1][ race_number ].laps ||
	    races[1][ race_number ].laps == 2 )
	{
	    if( number_of_dogs_entered < 9 &&
		dogs[ dogs_on_level[ order[i] ] ].status > 0 &&
		dogs[ dogs_on_level[ order[i] ] ].status < 20 )
	    {
		races[1][ race_number ].dogs[ number_of_dogs_entered ] =
		    &dogs[ dogs_on_level[ order[i] ] ];
		number_of_dogs_entered++;
	    }
	}
    }
        
    /* if there weren't enough dogs with the fitting preference,
       fill up the rest of the race with any kinds of dogs */
    for(i=0;i<number_of_dogs_on_level;i++)
    {
	if( entered_in_races(dogs[ dogs_on_level[ order[i] ] ]) == 0)
	{
	    if(number_of_dogs_entered < 9 &&
		dogs[ dogs_on_level[ order[i] ] ].status > 0 &&
		dogs[ dogs_on_level[ order[i] ] ].status < 20 )
	    {
		races[1][ race_number ].dogs[ number_of_dogs_entered ] =
		    &dogs[ dogs_on_level[ order[i] ] ];
		number_of_dogs_entered++;
	    }
	}
    }

    return 0;
}

/* number of laps the dog prefers in a race */
gint preference_of(dog Dog)
{
    if( sum(Dog.speed, 0, 98) > sum(Dog.stamina, 0, 98))
	return 1;
    else 
	return 3;
}

/* check whether a given dog is entered into one
   of the 12 races */
gint entered_in_races(dog Dog)
{
    gint i,j;

    for(i=0;i<12;i++)
	for(j=0;j<10;j++)
	    if(races[1][i].dogs[j] != NULL && Dog.id == races[1][i].dogs[j]->id)
		return 1;
    
    return 0;
}

/* decide the outcome of the 12 races and move
   races[1] (next races) to races[0] (last races) */
gint race_results(void)
{
    gint i,j;
    gint last_index;

    for(i=0;i<12;i++)
    {
	last_index = 10;
	
	/* see how many dogs participate in the race */
	for(j=9;j>=0;j--)
	    if(races[1][i].dogs[j] == NULL)
		last_index = j;
	
	/* compute race time for each dog */
	for(j=0;j<last_index;j++)
	    races[1][i].times[j] = 
		calculate_time(races[1][i].dogs[j], races[1][i].laps);

	/* find out whether there were injuries */
	calculate_injuries(&races[1][i], last_index);

	/* write the rankings based on the times */
	write_ranks(&races[1][i]);
	
	/* update the race statistics of the dogs */
	update_dog_statistics(races[1][i], last_index);
	
	/* move races[1] to races[0] */

	races[0][i].laps = races[1][i].laps;
	races[0][i].level = races[1][i].level;
	for(j=0;j<10;j++)
	{
	    races[0][i].times[j] = races[1][i].times[j];
	    races[0][i].dogs[j] = races[1][i].dogs[j];
	    races[0][i].ranks[j] = races[1][i].ranks[j];
	}
    }

    return 0;
}

/* sort the times from fastest to slowest.
   after sorting race.times[ ranks[0] ] is the fastest time
   and race.times[ ranks[9] ] the slowest.
   correspondingly  race.dogs[ ranks[0] ] the dog finishing first
   and race.dogs[ ranks[9] ] was last */
gint write_ranks(race *Race)
{
    gint i,j;
    gint ranks[10];

    for(i=0;i<10;i++)
	ranks[i] = 1;

    for(i=0;i<10;i++)
	for(j=i+1;j<10;j++)
	{
	    if(Race->times[i] < Race->times[j])
		ranks[j]++;
	    else
		ranks[i]++;
	}
    
    for(i=1;i<11;i++)
	for(j=0;j<10;j++)
	    if(ranks[j] == i)
		Race->ranks[i - 1] = j;
    
    return 0;
}

gint update_dog_statistics(race Race, gint last_index)
{
    gint i,j;

    /* make room for the new race result in
       'last_races' */
    for(i=0;i<last_index;i++)
    {
	for(j=9;j>1;j--)
	{
	    Race.dogs[i]->last_races[j] = Race.dogs[i]->last_races[j-2];
	}
    }
    for(i=0;i<last_index;i++)
    {
	/* increment number of races */
	Race.dogs[i]->races[0]++;
	
	/* if the dog finished top 5 note it */
 	if(i < 5)
	    Race.dogs[ Race.ranks[i] ]->races[i + 1]++;

	/* store for each dog its race result as the latest one
	   in 'last_races' */
	Race.dogs[ Race.ranks[i] ]->last_races[0] = i + 1;
	Race.dogs[i]->last_races[1] = Race.times[i];

	/* update BTS and BTE if necessary */
	if(Race.dogs[i]->best_times[Race.laps + 2] > Race.times[i])
	{
	    Race.dogs[i]->best_times[Race.laps + 2] = Race.times[i];
	    Race.dogs[i]->best_times[Race.laps - 1] = Race.times[i];
	}
	else if(Race.dogs[i]->best_times[Race.laps - 1] > Race.times[i])
	    Race.dogs[i]->best_times[Race.laps - 1] = Race.times[i];
    }

    return 0;
}

/* find out if any of the participating dogs has been injured.
   the probability of injury depends on fitness:
   at 0.6 it's 15% and at 1 its 0 */
gint calculate_injuries(race *Race, gint last_index)
{
    gint i,j;
    gfloat injury_duration_probabilities[8] =
	{0, 0.3, 0.5, 0.65, 0.8, 0.9, 0.95, 1};
    gfloat random;

    for(i=0;i<last_index;i++)
    {
	/* the right hand side is the probability of an injury
	   as a function of fitness */
	random = abs(1 - (gint)(Race->dogs[i]->fitness)) * 0.05;
	/* we exclude dogs that are becoming fathers from injury
	   so as not to confuse the 'father'-mark of status = 18 */
	if(rnd(0,1) < random && Race->dogs[i]->status != 18)
	{
	    /* store the event in the value below */
	    Race->times[i] = 110;
	    
	    /* calculate the duration of the injury (in semi-months);
	       as you can see above the probability of the
	       duration being long is small */
	    random = rnd(0,1);
	    for(j=0;j<7;j++)		
		if(injury_duration_probabilities[j] < random &&
		   random <= injury_duration_probabilities[j + 1])
		    
		    Race->dogs[i]->status = 21 + j;
	}
    }

    return 0;
}


/* write the races for the champions' trophy */
gint write_champions_trophy_races(void)
{
    gint i,j;

    for(i=0;i<12;i++)
    {
	if(i<6)
	{
	    /* we don't have levels here */
	    races[1][i].level = -1;
	    /* for each distance we have two races */
	    races[1][i].laps = (gint)((gfloat)i / 2) + 1;
	}
	
	for(j=0;j<10;j++)
	{
	    races[1][i].times[j] = 200;
	    races[1][i].dogs[j] = NULL;
	    races[1][i].ranks[j] = j;
	}
    }

    for(i=0;i<6;i++)
	if(qualified_kennels[i] != 0)
	    enter_kennel_in_trophy_races(qualified_kennels[i]);

    return 0;
}

/* choose the best six dogs of a kennel and
   enter them in the six trophy races */
gint enter_kennel_in_trophy_races(gint kennel_id)
{
    gint i;
    gint permutation[6];
    /* note the old ordering attribute
       as we will re-order the dogs to be able
       to choose the ones most suitable for the
       given race distances */
    gint old_sort_attribute = order[420];

    /* if the dogs are initially unsorted, we sort them
       by level afterwards */
    if( old_sort_attribute == 0 ||
	(abs(old_sort_attribute) > 7 &&
	 (old_sort_attribute / 10 < 2 || old_sort_attribute / 10 > 4) ) )
	old_sort_attribute = 3;


    /* write permutation of the races to avoid
       having the best dogs of each kennel in the 
       first few races */
    write_permutation(permutation, 0, 5);
   
    for(i=0;i<6;i++)
	enter_best_dog_in_race(kennel_id, permutation[i]);

    /* restore old sort order */
    sort_dogs(kennel_id, old_sort_attribute);
    
    return 0;
}

gint enter_best_dog_in_race(gint kennel_id, gint race_id)
{
    gint i, j;
    gint sort_attribute;
    
    if(races[1][race_id].laps == 1)
	sort_attribute = 20;
    else if(races[1][race_id].laps == 2)
	sort_attribute = 30;
    else
	sort_attribute = 40;

    /* sort the dogs in the kennel by their suitability
       for the given race distance */
    sort_dogs(kennel_id, sort_attribute);

    /* enter the best dog not yet entered */
    for(i = kennel_id * 20; i < (kennel_id + 1) * 20; i++)
    {
	if( entered_in_races(dogs[order[i]]) == 0)
	{
	    for(j=0;j<6;j++)
		if(races[1][race_id].dogs[j] == NULL)
		{
		    races[1][race_id].dogs[j] = &dogs[order[i]];
		    break;
		}
	    
	    break;
	}
    }
    
    return 0;
}

/* compute the points of a kennel in the trophy
   races up to race 'race_number'.
   for instance, if a kennel takes first place in the
   first trophy race and second in the second race,
   the points returned are 6 + 4 = 10 */
gint trophy_points_of_kennel(gint kennel_id, gint race_number)
{
    gint i,j;
    gint points = 0;

    for(i=0;i<race_number + 1;i++)
    {
	for(j=0;j<6;j++)
	{
	    if(races[0][j].dogs[races[0][i].ranks[j]] != NULL)
	    {
		if((gint)((gfloat)races[0][j].
			 dogs[races[0][i].ranks[j]]->id / 20)
		   == kennel_id)
		{
		    if(j == 0)
			points += 6;
		    else
			points += 5 - j;
		}
	    }
	}
    }

    return points;
}

/* write the six kennels that are qualified for
   the champions' trophy races into the array */
gint write_qualified_kennels(void)
{
    gint i,j, kennels_written = 2;
    gint temporary_order[20];
    gint best_kennels[3][6];

    for(i=0;i<6;i++)
	qualified_kennels[i] = -1;
    
    /* sort the kennels by the attributes 'wat',
       most w/r and p/r and note the first 6 */
    for(i=0;i<3;i++)
    {
	sort_kennel_statistics(i * 10 + 90, temporary_order, 0);
	for(j=0;j<6;j++)
	    best_kennels[i][j] = temporary_order[j];
    }
    
    /* find out which kennels are qualified.
       we consider the 'WAT' attribute most important,
       then points / race, then wins / race. */
    qualified_kennels[0] = best_kennels[0][0];
    qualified_kennels[1] = best_kennels[0][1];
    
    for(i=1;i<3;i++)
    {
	for(j=0;j<6;j++)
	{
	    if(is_in_list(qualified_kennels, 0, 5, best_kennels[i][j]) == 0)
	    {
		qualified_kennels[ kennels_written ] = best_kennels[i][j];
		kennels_written++;
	    }
	
	    if(kennels_written == i * 2 + 2)
		break;
	}
    }

    return 0;
}
